from .. import Check
import pandas as pd
from pathlib import Path
from typing import Dict


class BioChecks:
    def __init__(self, check: Check):
        self._check = check
        try:
            parent_path = Path(__file__).parent
            self._aminoacids = pd.read_csv(parent_path / "amino_acids.csv")
        except Exception:
            raise Exception("Unable to load aminoacid definitions")

    def is_dna(self, column: str, pct: float = 1.0, options: Dict[str, str] = {"name" : "is_dna"}):
        """Validates that a sequence contains only valid nucleotide bases of DNA strand"""
        self._check.has_pattern(column, r"^[GTCA]*$", pct, options=options)
        return self._check

    def is_protein(self, column: str, pct: float = 1.0, options: Dict[str, str] = {"name": "is_protein"}):
        """Verifies that a sequence contains only valid aminoacid 1-letter codes"""
        self._check.has_pattern(
            column, rf"^[{''.join(self._aminoacids['1_letter_code'].tolist())}]*$", pct, options=options
        )
        return self._check

    def is_cds(self, column: str, pct: float = 1.0, options: Dict[str, str] = {"name": "is_cds"}):
        """Verifies that a sequence contains the correct codons"""
        self._check.satisfies(
            column, f"({column} rlike '^ATG.*') and ({column} rlike '.*(TAA|TAG|TGA)$') and (length({column}) % 3 == 0)", pct, options=options
        )
        return self._check



import xml.etree.ElementTree as ET
import requests
from dataclasses import dataclass
import os
from operator import attrgetter as at
from functools import lru_cache
from typing import Dict


@dataclass
class Currency:
    country_name: str
    currency_name: str
    currency: str = None
    currency_number: str = None
    currency_units: str = None


@lru_cache
def _load_currencies():
    """External download from ISO website of currencies in XML format"""
    DEFAULT_ENDPOINT_4217 = "https://www.six-group.com/dam/download/financial-information/data-center/iso-currrency/lists/list-one.xml"
    response = requests.get(os.getenv("ISO_4217_ENDPOINT", DEFAULT_ENDPOINT_4217))
    xml = ET.fromstring(response.text.encode("utf-8"))

    def _get_ccy(element):
        currency_data = {tag.tag: tag.text.strip() for tag in list(element)}
        return at("currency")(Currency(*currency_data.values()))

    return set(filter(None, (map(_get_ccy, xml.findall("./CcyTbl/CcyNtry")))))


@lru_cache
def _load_countries():
    """External download from Google shared data of country codes and locations"""
    DEFAULT_ENDPOINT_3166 = "https://raw.githubusercontent.com/lukes/ISO-3166-Countries-with-Regional-Codes/master/all/all.xml"
    response = requests.get(os.getenv("ISO_4217_ENDPOINT", DEFAULT_ENDPOINT_3166))
    xml = ET.fromstring(response.text.encode("utf-8"))
    return list(map(lambda x: x.attrib["alpha-2"], xml.findall("./country")))


class ISO:
    """Abstraction of checks related to ISO standards"""

    CCY_CODE = "currency"
    CCY_NUMBER = "currency_number"
    COUNTRY_CODE = "country"
    COUNTRY_NAME = "name"

    def __init__(self, check):
        self._check = check
        self._ccy = []
        self._countries = []

    def iso_4217(self, column: str, pct: float = 1.0, options: Dict[str,str]={"name": "iso_4217"}):
        """It verifies a field against the international standard currency codes via code or number fields from ISO 4217"""
        self._ccy = _load_currencies()
        self._check.is_contained_in(column, self._ccy, pct, options=options)
        return self._check

    def iso_3166(self, column: str, pct: float = 1.0, options: Dict[str,str]={"name": "iso_3166"}):
        """Verifies that country codes are valid against the ISO standard 3166"""
        self._countries = _load_countries()
        self._check.is_contained_in(column, self._countries, pct, options=options)
        return self._check

    iso_currencies = iso_4217
    iso_countries = iso_3166



import os
import logging
import requests
from requests.exceptions import ConnectionError

logger = logging.getLogger("cuallee")
CUALLEE_CLOUD_HEADERS = {
    "Content-Type": "application/octet-stream",
    "Authorization": f"Bearer {os.getenv('CUALLEE_CLOUD_TOKEN')}",
}

try:
    import msgpack
except (ModuleNotFoundError, ImportError):
    logger.error("Module msgpack missing for cloud operations")


def standardize(check):
    return {
        "name": check.name,
        "date": check.date.strftime("%Y-%m-%d %H:%M:%S"),
        "level": check.level.name,
        "rows": check.rows,
        "rules": [
            {
                "ord": int(r.ordinal),
                "col": str(r.column),
                "met": str(r.method),
                "nam": str(r.name),
                "val": str(r.value),
                "vio": int(r.violations),
                "psr": float(r.pass_rate),
                "cov": float(r.coverage),
                "sta": str(r.status),
            }
            for r in check.rules
        ],
    }


def compress(check):
    return msgpack.packb(standardize(check))


def publish(check):
    """Send results to Cuallee Cloud"""
    try:
        requests.post(
            os.getenv("CUALLEE_CLOUD_HOST"),
            data=compress(check),
            headers=CUALLEE_CLOUD_HEADERS,
            verify=False,
        )
    except (ModuleNotFoundError, KeyError, ConnectionError) as error:
        logger.debug(f"Unable to send check to cuallee cloud: {str(error)}")
        pass



from typing import List, Tuple
from fpdf import FPDF

# from datetime import datetime, timezone


def pdf(data: List[Tuple[str]], name: str = "cuallee.pdf"):
    # today = datetime.now(timezone.utc).strftime("%Y%m%d%H%M%S")
    # style = FontFace(fill_color="#AAAAAA")
    pdf = FPDF(orientation="landscape", format="A4")
    pdf.add_page()
    pdf.set_font("Helvetica", size=6)
    irow = -1
    with pdf.table(
        borders_layout="SINGLE_TOP_LINE",
        width=270,
        col_widths=(15, 25, 25, 25, 115, 25, 25),
    ) as table:
        for data_row in data:
            row = table.row()
            for datum in data_row:
                row.cell(datum)
            irow += 1

    pdf.output(name)



import enum
import operator
import pandas as pd  # type: ignore
from dataclasses import dataclass
from typing import Dict, List, Union
from string import Template
from toolz import valfilter  # type: ignore
from google.cloud import bigquery
from cuallee import Check, ComputeEngine, Rule
import logging

logger = logging.getLogger("cuallee")


class ComputeMethod(enum.Enum):
    SQL = "SQL"
    TRANSFORM = "TRANSFORM"


@dataclass
class ComputeInstruction:
    predicate: Union[str, List[str], None]
    expression: str
    compute_method: ComputeMethod

    def __repr__(self):
        return f"ComputeInstruction({self.compute_method})"


class Compute(ComputeEngine):
    def __init__(self, table_name: str = None):
        self.compute_instruction: Union[ComputeInstruction, None] = None

    def _sum_predicate_to_integer(self, predicate) -> str:
        return f"SUM(CAST({predicate} AS INTEGER))".replace(",)", ")")

    def is_complete(self, rule: Rule):
        """Verify the absence of null values in a column"""
        predicate = f"{rule.column} IS NOT NULL"
        self.compute_instruction = ComputeInstruction(
            predicate,
            self._sum_predicate_to_integer(predicate),
            ComputeMethod.SQL,
        )
        return self.compute_instruction

    def is_empty(self, rule: Rule):
        """Verify the presence of null values in a column"""
        predicate = f"{rule.column} IS NULL"
        self.compute_instruction = ComputeInstruction(
            predicate,
            self._sum_predicate_to_integer(predicate),
            ComputeMethod.SQL,
        )
        return self.compute_instruction

    def are_complete(self, rule: Rule):
        """Verify the absence of null values in a column"""
        predicate = [f"{c} IS NOT NULL" for c in rule.column]
        self.compute_instruction = ComputeInstruction(
            predicate,
            "("
            + "+".join([self._sum_predicate_to_integer(p) for p in predicate])
            + f")/{len(rule.column)}",
            ComputeMethod.SQL,
        )
        return self.compute_instruction

    def is_unique(self, rule: Rule):
        """Validation for unique values in column"""
        predicate = None
        self.compute_instruction = ComputeInstruction(
            predicate,
            f"COUNT(DISTINCT {rule.column})",
            ComputeMethod.SQL,
        )
        return self.compute_instruction

    def has_cardinality(self, rule: Rule):
        """Validation of number of distinct values in column"""
        self.compute_instruction = ComputeInstruction(
            None,
            f"COUNT(DISTINCT({rule.column}))={rule.value}",
            ComputeMethod.SQL,
        )
        return self.compute_instruction

    def has_infogain(self, rule: Rule):
        """More than 1 distinct value"""
        self.compute_instruction = ComputeInstruction(
            None,
            f"COUNT(DISTINCT({rule.column})) > 1",
            ComputeMethod.SQL,
        )
        return self.compute_instruction

    def are_unique(self, rule: Rule):
        """Validation for unique values in a group of columns"""
        predicate = None
        self.compute_instruction = ComputeInstruction(
            predicate,
            "COUNT(DISTINCT CONCAT("
            + ", '_', ".join([f"{c}" for c in rule.column])
            + "))",
            ComputeMethod.SQL,
        )
        return self.compute_instruction

    def is_contained_in(self, rule: Rule):
        """Validation of column value in set of given values"""

        predicate = f"{rule.column} IN {rule.value}"
        self.compute_instruction = ComputeInstruction(
            predicate,
            self._sum_predicate_to_integer(predicate),
            ComputeMethod.SQL,
        )
        return self.compute_instruction

    def not_contained_in(self, rule: Rule):
        """Validation of column value not in a set of given values"""
        predicate = f"{rule.column} NOT IN {rule.value}"
        self.compute_instruction = ComputeInstruction(
            predicate,
            self._sum_predicate_to_integer(predicate),
            ComputeMethod.SQL,
        )
        return self.compute_instruction

    def is_daily(self, rule: Rule):
        """Validates that there is no missing dates using only week days in the date/timestamp column"""

        predicate = None

        def _execute(dataframe: bigquery.table.Table, key: str) -> str:
            day_mask = rule.value
            if not day_mask:
                day_mask = tuple([2, 3, 4, 5, 6])

            script = Template(
                """SELECT CASE WHEN numb_rows > 0 THEN CAST(numb_rows*-1 AS STRING) ELSE "true" END AS KEY$key FROM (SELECT COUNT (*) AS numb_rows FROM(SELECT full_interval.date FROM (SELECT date FROM (SELECT date, EXTRACT(DAYOFWEEK FROM date) AS dayofweek FROM UNNEST((SELECT GENERATE_DATE_ARRAY(min, max) AS date FROM (SELECT CAST(MIN($column) AS DATE) AS min, CAST(MAX($column) AS DATE)AS max FROM $table))) AS date ORDER BY date) WHERE dayofweek IN $value) AS full_interval LEFT OUTER JOIN $table AS table ON (full_interval.date=CAST(table.$column AS DATE)) WHERE $column IS NULL))""".strip()
            )

            return script.substitute(
                table=dataframe, key=key, column=rule.column, value=day_mask
            )

        self.compute_instruction = ComputeInstruction(
            predicate,
            _execute,
            ComputeMethod.TRANSFORM,
        )
        return self.compute_instruction


def _get_expressions(compute_set: Dict[str, ComputeInstruction]) -> str:
    """Get the expression for all the rules in check in one string"""

    return ", ".join(
        [
            compute_instruction.expression + f" AS KEY{key}"
            for key, compute_instruction in compute_set.items()
        ]
    )


def _build_query(expression_string: str, dataframe: bigquery.table.Table) -> str:
    """Build query final query"""

    return f"SELECT {expression_string} FROM `{str(dataframe)}`"


def _compute_query_method(
    client, dataframe: bigquery.table.Table, compute_set: Dict[str, ComputeInstruction]
) -> Dict:
    """Compute rules throught query"""

    # Filter expression directed to sql
    _sql = lambda x: x.compute_method.name == ComputeMethod.SQL.name
    sql_set = valfilter(_sql, compute_set)

    if sql_set:
        expression_string = _get_expressions(sql_set)
        query = _build_query(expression_string, dataframe)
        logger.error(query)

        return client.query(query).to_arrow().to_pandas().to_dict(orient="records")[0]
    else:
        return {}


def _compute_transform_method(
    client, dataframe: bigquery.table.Table, compute_set: Dict[str, ComputeInstruction]
) -> Dict:
    """Compute rules that require to pass the table as variable"""

    # Filter expression directed to transform
    _transform = lambda x: x.compute_method.name == ComputeMethod.TRANSFORM.name
    transform_set = valfilter(_transform, compute_set)

    return {
        f"KEY{k}": operator.itemgetter(f"KEY{k}")(
            client.query(compute_instruction.expression(dataframe, k))
            .to_arrow()
            .to_pandas()
            .to_dict(orient="records")[0]
        )
        for k, compute_instruction in transform_set.items()
    }


def _compute_row(client, dataframe: bigquery.table.Table) -> Dict:
    """Get the number of rows"""

    return (
        client.query(f"SELECT COUNT(*) AS count FROM `{str(dataframe)}`")
        .to_arrow()
        .to_pandas()
        .to_dict(orient="records")
    )


def _calculate_violations(result, nrows) -> Union[int, float]:
    """Return the number of violations for each rule"""

    if (result == "true") or (result is True):
        return 0
    elif (result == "false") or (result is False):
        return nrows
    elif int(result) < 0:
        return abs(int(result))
    else:
        return nrows - int(result)


def _calculate_pass_rate(result, nrows) -> float:
    """Return the pass rate for each rule"""

    if (result == "true") or (result is True):
        return 1.0
    elif (result == "false") or (result is False):
        return 0.0
    elif int(result) < 0:
        if abs(int(result)) < nrows:
            return 1 - (abs(int(result)) / nrows)
        elif abs(int(result)) == nrows:
            return 0.5
        else:
            return nrows / abs(int(result))
    else:
        return int(result) / nrows


def _evaluate_status(pass_rate, pass_threshold) -> str:
    """Return the status for each rule"""

    if pass_rate >= pass_threshold:
        return "PASS"
    else:
        return "FAIL"


def validate_data_types(rules: List[Rule], dataframe: str) -> bool:
    """Validate the datatype of each column according to the CheckDataType of the rule's method"""
    return True


def compute(rules: Dict[str, Rule]) -> Dict:
    """Create dictionnary containing compute instruction for each rule."""
    return {k: operator.methodcaller(v.method, v)(Compute()) for k, v in rules.items()}


def summary(check: Check, dataframe: bigquery.table.Table):
    """Compute all rules in this check from table loaded in BigQuery"""

    # Check that user is connected to BigQuery
    try:
        client = bigquery.Client()
    except Exception as error:
        print(
            f"You are not connected to the BigQuery cloud. Please verify the steps followed during the Authenticate API requests step. {str(error)}"
        )

    # Compute the expression
    computed_expressions = compute(check._rule)

    query_expression = _compute_query_method(client, dataframe, computed_expressions)
    query_transform = _compute_transform_method(client, dataframe, computed_expressions)
    query_result = {**query_expression, **query_transform}

    # Compute the total number of rows
    rows = _compute_row(client, dataframe)[0]["count"]

    # Results
    computation_basis = [
        {
            "id": index,
            "timestamp": check.date.strftime("%Y-%m-%d %H:%M:%S"),
            "check": check.name,
            "level": check.level.name,
            "column": rule.column,
            "rule": rule.name,
            "value": rule.value,
            "rows": rows,
            "violations": _calculate_violations(query_result[f"KEY{hash_key}"], rows),
            "pass_rate": _calculate_pass_rate(query_result[f"KEY{hash_key}"], rows),
            "pass_threshold": rule.coverage,
            "status": _evaluate_status(
                _calculate_pass_rate(query_result[f"KEY{hash_key}"], rows),
                rule.coverage,
            ),
        }
        for index, (hash_key, rule) in enumerate(check._rule.items(), 1)
    ]

    return pd.DataFrame(computation_basis).set_index("id")



import daft
import operator
import statistics
import numpy as np
import pandas as pd

from math import log2
from toolz import first
from typing import Union
from numbers import Number
from typing import Dict, List

from itertools import groupby
from operator import itemgetter

from cuallee import Check, Rule


class Compute:

    def is_complete(self, rule: Rule, dataframe: daft.DataFrame) -> Union[bool, int]:
        col_name = daft.col(rule.column)
        perdicate = col_name.not_null().cast(daft.DataType.int64())
        return dataframe.select(perdicate).sum(col_name).to_pandas().iloc[0, 0]

    def is_empty(self, rule: Rule, dataframe: daft.DataFrame) -> Union[bool, int]:
        col_name = daft.col(rule.column)
        perdicate = col_name.is_null().cast(daft.DataType.int64())
        return dataframe.select(perdicate).sum(col_name).to_pandas().iloc[0, 0]

    def are_complete(self, rule: Rule, dataframe: daft.DataFrame) -> Union[bool, int]:
        col_names = rule.column
        perdicate = [
            daft.col(col_name).not_null().cast(daft.DataType.int64())
            for col_name in col_names
        ]
        col_names_list = [daft.col(col_name) for col_name in rule.column]
        return dataframe.select(*perdicate).sum(col_names_list).to_pandas().astype(
            int
        ).sum().sum() / len(col_names)

    def is_unique(self, rule: Rule, dataframe: daft.DataFrame) -> Union[bool, int]:
        perdicate = daft.col(rule.column)
        return (
            dataframe.select(perdicate)
            .distinct()
            .count(perdicate)
            .to_pandas()
            .iloc[0, 0]
        )

    def are_unique(self, rule: Rule, dataframe: daft.DataFrame) -> Union[bool, int]:
        # TODO: Find a way to do this in daft and not pandas
        perdicate = [daft.col(col_name) for col_name in rule.column]
        return dataframe.select(*perdicate).to_pandas().nunique().sum() / len(
            rule.column
        )

    def is_greater_than(
        self, rule: Rule, dataframe: daft.DataFrame
    ) -> Union[bool, int]:
        col_name = daft.col(rule.column)
        perdicate = (col_name > rule.value).cast(daft.DataType.int64())
        return dataframe.select(perdicate).sum(col_name).to_pandas().iloc[0, 0]

    def is_greater_or_equal_than(
        self, rule: Rule, dataframe: daft.DataFrame
    ) -> Union[bool, int]:
        col_name = daft.col(rule.column)
        perdicate = (col_name >= rule.value).cast(daft.DataType.int64())
        return dataframe.select(perdicate).sum(col_name).to_pandas().iloc[0, 0]

    def is_less_than(self, rule: Rule, dataframe: daft.DataFrame) -> Union[bool, int]:
        col_name = daft.col(rule.column)
        perdicate = (col_name < rule.value).cast(daft.DataType.int64())
        return dataframe.select(perdicate).sum(col_name).to_pandas().iloc[0, 0]

    def is_less_or_equal_than(
        self, rule: Rule, dataframe: daft.DataFrame
    ) -> Union[bool, int]:
        col_name = daft.col(rule.column)
        perdicate = (col_name <= rule.value).cast(daft.DataType.int64())
        return dataframe.select(perdicate).sum(col_name).to_pandas().iloc[0, 0]

    def is_equal_than(self, rule: Rule, dataframe: daft.DataFrame) -> Union[bool, int]:
        col_name = daft.col(rule.column)
        perdicate = (col_name == rule.value).cast(daft.DataType.int64())
        return dataframe.select(perdicate).sum(col_name).to_pandas().iloc[0, 0]

    def has_pattern(self, rule: Rule, dataframe: daft.DataFrame) -> Union[bool, int]:
        col_name = daft.col(rule.column)
        perdicate = (col_name.str.match(rule.value)).cast(daft.DataType.int64())
        return dataframe.select(perdicate).sum(col_name).to_pandas().iloc[0, 0]

    def has_min(self, rule: Rule, dataframe: daft.DataFrame) -> Union[bool, int]:
        col_name = daft.col(rule.column)
        return (
            dataframe.agg(col_name.min())
            .select(col_name == rule.value)
            .to_pandas()
            .iloc[0, 0]
        )

    def has_max(self, rule: Rule, dataframe: daft.DataFrame) -> Union[bool, int]:
        col_name = daft.col(rule.column)
        return (
            dataframe.agg(col_name.max())
            .select(col_name == rule.value)
            .to_pandas()
            .iloc[0, 0]
        )

    def has_std(self, rule: Rule, dataframe: daft.DataFrame) -> Union[bool, int]:

        @daft.udf(return_dtype=daft.DataType.float64())
        def std_dev(data):
            return [statistics.stdev(data.to_pylist())]

        return (
            dataframe.select(std_dev(daft.col(rule.column))).to_pandas().iloc[0, 0]
            == rule.value
        )

    def has_mean(self, rule: Rule, dataframe: daft.DataFrame) -> Union[bool, int]:
        col_name = daft.col(rule.column)
        return (
            dataframe.agg(col_name.mean())
            .select(col_name == rule.value)
            .to_pandas()
            .iloc[0, 0]
        )

    def has_sum(self, rule: Rule, dataframe: daft.DataFrame) -> Union[bool, int]:
        perdicate = daft.col(rule.column)
        return (
            dataframe.select(perdicate).sum(perdicate).to_pandas().iloc[0, 0]
            == rule.value
        )

    def has_cardinality(
        self, rule: Rule, dataframe: daft.DataFrame
    ) -> Union[bool, int]:
        perdicate = daft.col(rule.column)
        return (
            dataframe.select(perdicate)
            .distinct()
            .count(perdicate)
            .to_pandas()
            .iloc[0, 0]
            == rule.value
        )

    def has_infogain(self, rule: Rule, dataframe: daft.DataFrame) -> Union[bool, int]:
        perdicate = daft.col(rule.column)
        return (
            dataframe.select(perdicate)
            .distinct()
            .count(perdicate)
            .to_pandas()
            .iloc[0, 0]
            > 1
        )

    def is_between(self, rule: Rule, dataframe: daft.DataFrame) -> Union[bool, int]:
        col_name = daft.col(rule.column)
        perdicate = (
            (col_name >= min(rule.value)).__and__(col_name <= max(rule.value))
        ).cast(daft.DataType.int64())
        return dataframe.select(perdicate).sum(col_name).to_pandas().iloc[0, 0]

    def is_contained_in(
        self, rule: Rule, dataframe: daft.DataFrame
    ) -> Union[bool, int]:
        rule_value = list(rule.value) if isinstance(rule.value, tuple) else rule.value
        col_name = daft.col(rule.column)
        perdicate = col_name.is_in(rule_value).cast(daft.DataType.int64())
        return dataframe.select(perdicate).sum(col_name).to_pandas().iloc[0, 0]

    def not_contained_in(
        self, rule: Rule, dataframe: daft.DataFrame
    ) -> Union[bool, int]:
        rule_value = list(rule.value) if isinstance(rule.value, tuple) else rule.value
        col_name = daft.col(rule.column)
        perdicate = col_name.is_in(rule_value).__ne__(True).cast(daft.DataType.int64())
        return dataframe.select(perdicate).sum(col_name).to_pandas().iloc[0, 0]

    def has_percentile(self, rule: Rule, dataframe: daft.DataFrame) -> Union[bool, int]:
        # TODO: Find a way to do this in daft and not rely on pandas
        perdicate = daft.col(rule.column)
        return (
            np.percentile(dataframe.select(perdicate).to_pandas().values, rule.settings["percentile"] * 100)  # type: ignore
            == rule.value  # type: ignore
        )

    def has_max_by(self, rule: Rule, dataframe: daft.DataFrame) -> Union[bool, int]:
        predicate_1 = daft.col("id")

        max_value = dataframe.agg(daft.col("id2").max()).to_pandas().iloc[0, 0]
        predicate_2 = daft.col("id2").__eq__(max_value)

        return (
            dataframe.where(predicate_2).select(predicate_1).to_pandas().iloc[0, 0]
            == rule.value
        )

    def has_min_by(self, rule: Rule, dataframe: daft.DataFrame) -> Union[bool, int]:
        predicate_1 = daft.col("id")

        min_value = dataframe.agg(daft.col("id2").min()).to_pandas().iloc[0, 0]
        predicate_2 = daft.col("id2").__eq__(min_value)

        return (
            dataframe.where(predicate_2).select(predicate_1).to_pandas().iloc[0, 0]
            == rule.value
        )

    def has_correlation(
        self, rule: Rule, dataframe: daft.DataFrame
    ) -> Union[bool, int]:

        @daft.udf(return_dtype=daft.DataType.float64())
        def correlation(x, y):
            return [statistics.correlation(x.to_pylist(), y.to_pylist())]

        return (
            dataframe.select(
                correlation(
                    daft.col(rule.column[0])
                    .is_null()
                    .if_else(daft.lit(0), daft.col(rule.column[0]))
                    .alias("x"),
                    daft.col(rule.column[1])
                    .is_null()
                    .if_else(daft.lit(0), daft.col(rule.column[1]))
                    .alias("y"),
                )
            )
            .to_pandas()
            .iloc[0, 0]
            == rule.value
        )

    def satisfies(self, rule: Rule, dataframe: daft.DataFrame) -> Union[bool, int]:
        """
        Note: At the moment `daft.DataFrame.where` just accepts Expression not string
        """

        if not isinstance(rule.value, daft.Expression):
            raise ValueError("The value of the rule must be an `daft.Expression`")

        if isinstance(rule.column, list) or isinstance(rule.column, tuple):
            cols = [daft.col(col) for col in rule.column]
            return dataframe.where(rule.value).count(*cols).to_pandas().iloc[0, 0]
        else:
            col = daft.col(rule.column)
            return dataframe.where(rule.value).count(col).to_pandas().iloc[0, 0]

    def has_entropy(self, rule: Rule, dataframe: daft.DataFrame) -> Union[bool, int]:

        @daft.udf(return_dtype=daft.DataType.float64())
        def entropy(probabilities: daft.Series):
            return [-sum(p * log2(p) for p in probabilities.to_pylist() if p > 0)]

        # Step 1: Cast the column to string and perform groupby and count
        dataframe = (
            dataframe.with_column(
                "label", daft.col(rule.column).cast(daft.DataType.string())
            )
            .groupby("label")
            .count(rule.column)
        )

        # Step 2: Calculate the total sum of counts separately
        total_count = int(
            dataframe.select(daft.col(rule.column)).to_pandas()[rule.column].sum()
        )

        # Step 3: Calculate the probabilities column using the total sum of counts
        dataframe = dataframe.with_column(
            "probabilities", daft.col(rule.column) / total_count
        )

        return dataframe.select(entropy(daft.col("probabilities"))).to_pandas().iloc[
            0, 0
        ] == float(rule.value)

    def is_on_weekday(self, rule: Rule, dataframe: daft.DataFrame) -> Union[bool, int]:
        col_name = daft.col(rule.column)
        perdicate = (
            col_name.dt.day_of_week().is_in([0, 1, 2, 3, 4]).cast(daft.DataType.int64())
        )
        return dataframe.select(perdicate).sum(col_name).to_pandas().iloc[0, 0]

    def is_on_weekend(self, rule: Rule, dataframe: daft.DataFrame) -> Union[bool, int]:
        col_name = daft.col(rule.column)
        perdicate = col_name.dt.day_of_week().is_in([5, 6]).cast(daft.DataType.int64())
        return dataframe.select(perdicate).sum(col_name).to_pandas().iloc[0, 0]

    def is_on_monday(self, rule: Rule, dataframe: daft.DataFrame) -> Union[bool, int]:
        col_name = daft.col(rule.column)
        perdicate = col_name.dt.day_of_week().__eq__(0).cast(daft.DataType.int64())
        return dataframe.select(perdicate).sum(col_name).to_pandas().iloc[0, 0]

    def is_on_tuesday(self, rule: Rule, dataframe: daft.DataFrame) -> Union[bool, int]:
        col_name = daft.col(rule.column)
        perdicate = col_name.dt.day_of_week().__eq__(1).cast(daft.DataType.int64())
        return dataframe.select(perdicate).sum(col_name).to_pandas().iloc[0, 0]

    def is_on_wednesday(
        self, rule: Rule, dataframe: daft.DataFrame
    ) -> Union[bool, int]:
        col_name = daft.col(rule.column)
        perdicate = col_name.dt.day_of_week().__eq__(2).cast(daft.DataType.int64())
        return dataframe.select(perdicate).sum(col_name).to_pandas().iloc[0, 0]

    def is_on_thursday(self, rule: Rule, dataframe: daft.DataFrame) -> Union[bool, int]:
        col_name = daft.col(rule.column)
        perdicate = col_name.dt.day_of_week().__eq__(3).cast(daft.DataType.int64())
        return dataframe.select(perdicate).sum(col_name).to_pandas().iloc[0, 0]

    def is_on_friday(self, rule: Rule, dataframe: daft.DataFrame) -> Union[bool, int]:
        col_name = daft.col(rule.column)
        perdicate = col_name.dt.day_of_week().__eq__(4).cast(daft.DataType.int64())
        return dataframe.select(perdicate).sum(col_name).to_pandas().iloc[0, 0]

    def is_on_saturday(self, rule: Rule, dataframe: daft.DataFrame) -> Union[bool, int]:
        col_name = daft.col(rule.column)
        perdicate = col_name.dt.day_of_week().__eq__(5).cast(daft.DataType.int64())
        return dataframe.select(perdicate).sum(col_name).to_pandas().iloc[0, 0]

    def is_on_sunday(self, rule: Rule, dataframe: daft.DataFrame) -> Union[bool, int]:
        col_name = daft.col(rule.column)
        perdicate = col_name.dt.day_of_week().__eq__(6).cast(daft.DataType.int64())
        return dataframe.select(perdicate).sum(col_name).to_pandas().iloc[0, 0]

    def is_on_schedule(self, rule: Rule, dataframe: daft.DataFrame) -> Union[bool, int]:
        col_name = daft.col(rule.column)
        perdicate = (
            (col_name.dt.hour() >= min(rule.value)).__and__(
                col_name.dt.hour() <= max(rule.value)
            )
        ).cast(daft.DataType.int64())
        return dataframe.select(perdicate).sum(col_name).to_pandas().iloc[0, 0]

    def is_daily(self, rule: Rule, dataframe: daft.DataFrame) -> complex:
        if rule.value is None:
            day_mask = [0, 1, 2, 3, 4]
        else:
            day_mask = rule.value

        agg_predicate = [
            daft.col(rule.column).dt.date().min().alias("min"),
            daft.col(rule.column).dt.date().max().alias("max"),
        ]
        select_predicate = [
            daft.col("min").cast(daft.DataType.string()),
            daft.col("max").cast(daft.DataType.string()),
        ]

        lower, upper = (
            dataframe.agg(*agg_predicate)
            .select(*select_predicate)
            .to_pandas()
            .iloc[0]
            .tolist()
        )

        sequence = (
            pd.date_range(start=lower, end=upper, freq="D").rename("ts").to_frame()
        )
        sequence = list(
            sequence[sequence.ts.dt.dayofweek.isin(day_mask)]
            .reset_index(drop=True)
            .ts.unique()
            .astype("datetime64[ms]")
        )

        filter_predicate = daft.col(rule.column).dt.day_of_week().is_in(day_mask)
        select_predicate = daft.col(rule.column).cast(
            daft.DataType.from_numpy_dtype("datetime64[ms]")
        )
        delivery = (
            dataframe.where(filter_predicate)
            .select(select_predicate)
            .to_pandas()[rule.column]
            .tolist()
        )

        # No difference between sequence of daily as a complex number
        return complex(len(dataframe), len(set(sequence).difference(delivery)))

    def is_inside_interquartile_range(
        self, rule: Rule, dataframe: daft.DataFrame
    ) -> Union[bool, complex]:
        col_name = daft.col(rule.column)
        lower, upper = (
            dataframe.select(col_name).to_pandas().quantile(rule.value).values
        )
        perdicate = ((col_name >= lower).__and__(col_name <= upper)).cast(
            daft.DataType.int64()
        )
        return dataframe.select(perdicate).sum(col_name).to_pandas().iloc[0, 0]

    def has_workflow(self, rule: Rule, dataframe: daft.DataFrame) -> Union[bool, int]:

        next_event = "CUALLEE_NEXT_EVENT"
        cuallee_edge = "CUALLEE_EDGE"
        group, event, order = rule.column

        edges = [f"({a}, {b})" for a, b in rule.value]

        @daft.udf(return_dtype=daft.DataType.string())
        def create_edge(x: daft.Series, y: daft.Series):
            return [f"({a}, {b})" for a, b in zip(x.to_pylist(), y.to_pylist())]

        @daft.udf(return_dtype=daft.DataType.string())
        def calculate_next_event(group, order, event):

            # Step 1: Zip the lists together
            zipped_data = list(
                zip(event.to_pylist(), group.to_pylist(), order.to_pylist())
            )

            # Step 2: Sort by group, then by order
            sorted_data = sorted(zipped_data, key=itemgetter(1, 2))

            # Function to calculate the next event within each group
            def calculate_next_event(group):
                items = list(group)
                for i in range(len(items)):
                    # Assign the next event's data if it's not the last item; otherwise, assign None
                    next_event = items[i + 1][0] if i + 1 < len(items) else None
                    yield items[i] + (next_event,)

            # Step 3 & 4: Group by 'group' and calculate the next event
            grouped_data = groupby(sorted_data, key=itemgetter(1))
            result = [
                item
                for _, group in grouped_data
                for item in calculate_next_event(group)
            ]

            _, _, _, next_events = zip(*result)

            return [str(item) for item in next_events]

        return (
            (
                dataframe.with_column(
                    next_event,
                    calculate_next_event(
                        daft.col(group), daft.col(order), daft.col(event)
                    ),
                )
                .with_column(
                    cuallee_edge, create_edge(daft.col(event), daft.col(next_event))
                )
                .select(
                    (daft.col(cuallee_edge).is_in(edges))
                    .alias("key")
                    .cast(daft.DataType.int64())
                )
                .sum(daft.col("key"))
            )
            .to_pandas()
            .iloc[0, 0]
        )


def compute(rules: Dict[str, Rule]):
    """Daft computes directly on the predicates"""
    return True


# TODO: Implement validate_data_types for daft
def validate_data_types(rules: List[Rule], dataframe: daft.DataFrame):
    """Validate the datatype of each column according to the CheckDataType of the rule's method"""
    return True


def summary(check: Check, dataframe: daft.DataFrame):
    compute = Compute()
    unified_results = {
        rule.key: [operator.methodcaller(rule.method, rule, dataframe)(compute)]
        for rule in check.rules
    }

    def _calculate_violations(result, nrows):
        if isinstance(result, (bool, np.bool_)):
            if result:
                return 0
            else:
                return nrows
        elif isinstance(result, Number):
            if isinstance(result, complex):
                return result.imag
            else:
                return nrows - result

    def _calculate_pass_rate(result, nrows):
        if isinstance(result, (bool, np.bool_)):
            if result:
                return 1.0
            else:
                return 0.0
        elif isinstance(result, Number):
            if isinstance(result, complex):
                if result.imag > 0:
                    if result.imag > nrows:
                        return nrows / result.imag
                    else:
                        return result.imag / nrows
                else:
                    return 1.0

            else:
                return result / nrows

    def _evaluate_status(pass_rate, pass_threshold):
        if pass_rate >= pass_threshold:
            return "PASS"

        return "FAIL"

    rows = len(dataframe)

    computation_basis = [
        {
            "id": index,
            "timestamp": check.date.strftime("%Y-%m-%d %H:%M:%S"),
            "check": check.name,
            "level": check.level.name,
            "column": rule.column,
            "rule": rule.name,
            "value": rule.value,
            "rows": rows,
            "violations": _calculate_violations(first(unified_results[hash_key]), rows),
            "pass_rate": _calculate_pass_rate(first(unified_results[hash_key]), rows),
            "pass_threshold": rule.coverage,
            "status": _evaluate_status(
                _calculate_pass_rate(first(unified_results[hash_key]), rows),
                rule.coverage,
            ),
        }
        for index, (hash_key, rule) in enumerate(check._rule.items(), 1)
    ]
    return daft.from_pylist(computation_basis)
